#include "mdlexport.h"
static vec3_t anorms_table[162] = {
#include "anorms.h"
};


void writedummyskin(std::ofstream& mdlfile) //export a dummy skin
{
	int32_t skintype = 0;
	int8_t check[17] = "aaaaaaaaZZZZZZZZ";//make a checkerboard pattern
	mdlfile.write( (char*)&skintype ,4);
	for(int32_t i = 0; i < (256*16) ; i++)
		mdlfile.write((char*)check,16);
}

static int32_t q1_skinheight = 256;
static int32_t q1_skinwidth = 256;


uint8_t findnormal (q3_vertex v)
{
	float theta = (float)(v.norm >> 8) * 2 * M_PI / 255.0f;
	float phi = (float)(v.norm & 255) * 2 * M_PI / 255.0f;
	
	vec3_t normal;
	if(v.norm == 0)
	{
		normal[0] = 0;
		normal[1] = 0;
		normal[2] = 1;
	}
	else if(v.norm == 32768)
	{
		normal[0] = 0;
		normal[1] = 0;
		normal[2] = -1;
	}
	else
	{
		normal[0] = cos( theta) * sin (phi);
		normal[1] = sin( theta) * sin (phi);
		normal[2] = cos (phi);
	}

	uint8_t best;
	float bestproduct;
	best = 0;
	bestproduct = anorms_table[0][0] * normal[0] + anorms_table[0][1] * normal[1] + anorms_table[0][2] * normal[2];
	for(int32_t i = 1; i < 162; ++i)
	{
		float currentproduct = anorms_table[i][0] * normal[0] + anorms_table[i][1] * normal[1] + anorms_table[i][2] * normal[2];
		if(currentproduct > bestproduct)
		{
			bestproduct = currentproduct;
			best = (unsigned char) i;
		}
	}
	return best;
}




void findbounds(float* mins, float* maxs, float& rad, q3_model& q3mdl)
{
	mins[0] = q3mdl.frames[0].mins[0];
	mins[1] = q3mdl.frames[0].mins[1];
	mins[2] = q3mdl.frames[0].mins[2];

	maxs[0] = q3mdl.frames[0].maxs[0];
	maxs[1] = q3mdl.frames[0].maxs[1];
	maxs[2] = q3mdl.frames[0].maxs[2];

	rad = 0;
	for each (q3_frame f in q3mdl.frames)
	{
		if(f.mins[0] < mins[0])
			mins[0] = f.mins[0];
		if(f.mins[1] < mins[1])
			mins[1] = f.mins[1];
		if(f.mins[2] < mins[2])
			mins[2] = f.mins[2];

		if(f.maxs[0] > maxs[0])
			maxs[0] = f.maxs[0];
		if(f.maxs[1] > maxs[1])
			maxs[1] = f.maxs[1];
		if(f.maxs[2] > maxs[2])
			maxs[2] = f.maxs[2];

		if(f.radius > rad)
			rad = f.radius;
	}
}

void export_to_mdl(std::ofstream& mdlfile, q3_model& q3mdl, skin& q1skin, options settings)
{
	float mins[3];
	float maxs[3];
	q1_header head;
	findbounds(mins, maxs, head.boundingradius, q3mdl);

	head.scale[0] = (maxs[0] - mins[0]) / 255.0f;
	head.scale[1] = (maxs[1] - mins[1]) / 255.0f;
	head.scale[2] = (maxs[2] - mins[2]) / 255.0f;

	head.translate[0] = mins[0];
	head.translate[1] = mins[1];
	head.translate[2] = mins[2]; //head.radius is added to the header directly by findbounds

	head.ident = 1330660425; //move to constructor?
	head.version = 6;

	head.eyeposition[0] = head.eyeposition[1] = 0;
	head.eyeposition[2] = mins[2];

	head.num_skins = 1;
	if(!q1skin.height)
	{
		head.skinheight = q1_skinheight; //magic!
		head.skinwidth = q1_skinwidth;
	}
	else
	{
		head.skinheight = q1skin.height; //magic!
		head.skinwidth = q1skin.width;
	}
	int32_t vertcount = 0;
	int32_t tricount = 0;

	std::vector<q3_surface>::iterator surfiter;
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		vertcount += (*surfiter).header.num_verts;
		tricount += (*surfiter).header.num_triangles;
	}
	head.num_verts = vertcount;
	head.num_tris = tricount;
	head.num_frames = (int)q3mdl.frames.size();

	head.synctype = 1; //almost always preferable
	head.flags = 0;
	head.size = 1; //arbitrary

	mdlfile.write((char*)&head, sizeof(head));

	//skin
	if(!q1skin.height)
		writedummyskin(mdlfile);
	else
	{
		std::cout << "writing pcx skin, width: " << q1skin.width <<", height: " << q1skin.height << std::endl;
		if(!q1skin.skindata)
			std::cout << "skindata corrupt" << std::endl;
		int32_t skintype = 0;
		mdlfile.write( (char*)&skintype ,4);
		for(int32_t i = 0; i < (q1skin.height * q1skin.width) ; ++i)
			mdlfile.write(q1skin.skindata+i, 1);
	}

	//uv coords
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		int32_t type = 0;
		for each(q3_texcoord uv in (*surfiter).texcoord)
		{
			int32_t u = (int)(uv.u *head.skinwidth);
			int32_t v = (int)(uv.v *head.skinheight);

			u = std::max<int32_t>(0,std::min(u,head.skinwidth)); //clamping to range
			v = std::max<int32_t>(0,std::min(v,head.skinheight));
			
			mdlfile.write((char*)&type, 4);
			mdlfile.write((char*)&u, 4);
			mdlfile.write((char*)&v, 4);
		}
	}

	//triangles
	int32_t triindexbump = 0; //increase the index of triangle vertices in later objects
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		int32_t type = 1;
		for each(q3_triangle tri in (*surfiter).triangle)
		{
			q3_triangle temptri = tri;
			temptri.index[0] += triindexbump;
			temptri.index[1] += triindexbump;
			temptri.index[2] += triindexbump;
			mdlfile.write((char*)&type, 4);
			mdlfile.write((char*)&temptri.index[0], 4);
			mdlfile.write((char*)&temptri.index[1], 4);
			mdlfile.write((char*)&temptri.index[2], 4);
		}
		triindexbump += (*surfiter).header.num_verts;
	}

	//frames
	int32_t type = 0;
	uint8_t maxcomp[4] = {255, 255, 255, 0}; //compressed bounding box format
	uint8_t mincomp[4] = {0, 0, 0, 0}; //this is actually a cheat which may be fixed later
	int32_t i = 0;
	uint8_t coords[4];
	for each(q3_frame f in q3mdl.frames)
	{
		mdlfile.write((char*)&type, 4);
		mdlfile.write((char*)&mincomp, 4);
		mdlfile.write((char*)&maxcomp, 4);
		mdlfile.write((char*)f.name, 16);
		
		for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
		{
			for each (q3_vertex v in (*surfiter).frame.at(i).vertex)
			{
				coords[0] = (unsigned char) (0.5+(v.x * MD3_XYZ_SCALE - mins[0]) / (head.scale[0]));
				coords[1] = (unsigned char) (0.5+(v.y * MD3_XYZ_SCALE - mins[1]) / (head.scale[1]));
				coords[2] = (unsigned char) (0.5+(v.z * MD3_XYZ_SCALE - mins[2]) / (head.scale[2]));
				coords[3] = findnormal(v);
				mdlfile.write((char*)&coords, 4);
			}
		}
		i++;
	}
}
